// Copyright 1992-1998 by Jon Dart.  All Rights Reserved.

#include "stdafx.h"
#include "book.h"
#include "movegen.h"
#include "moveord.h"
#include "globals.h"
#include "util.h"
extern "C"
{
#include <stdlib.h>
#include <time.h>
};

// Number of moves out of book at which learning occurs:
#define LEARNING_HORIZON 5
// Score threshold at which learning occurs
#define LEARNING_SCORE_THRESHOLD (Constants::PawnValue/3)
// Minimum search depth for learning
#define LEARNING_MIN_DEPTH 6

Book::Book()
{
   time_t t;
   srand((unsigned)time(&t));
   reader = new Book_Reader(MAIN_BOOK_NAME,MAIN_MAPPING_NAME);
   style_reader = new Book_Reader(STYLE_BOOK_NAME,STYLE_MAPPING_NAME);
}       

Book::~Book()
{
   delete reader;
   delete style_reader;
   Book_Entry::freeAll(TRUE);
}

Move Book::book_move( const Board &b, enum styles style,
                      Book_Info &info)
{
   // First try to find a move matching the preferred style
   // in the style book.
   info.clear();
   if ((style != Balanced) && style_reader->is_open())
   {
      Book_Location loc;
      style_reader->head(b, style, loc);
      Move m = style_reader->pick(b, loc, info);
      if (!IsNull(m))
         return m;
   }
   // No style move, try to find a move in the regular opening
   // book.
   if (!reader->is_open())
   {
      return NullMove;
   }
   else
   {
      Book_Location loc;
      reader->head(b,Balanced,loc);
      return reader->pick(b, loc, info);
   }
   return NullMove;
}

unsigned Book::book_moves( const Board &b, Move *moves, 
                           const unsigned limit)
{
   if (!reader->is_open())
      return 0;
   return reader->book_moves(b,moves,limit);

}

void Book::learn()
{
   int count = LEARNING_HORIZON;
   int last_book_move = -1;
   int min_score = Constants::BIG, max_score = -Constants::BIG;
   int i;
   for (i = theLog->current()-1; i>=0 && count>=0; i-=2,--count)
   {
      const Log_Entry &entry = (*theLog)[i];
      const Book_Info &book_info = entry.get_book_info();
      if (book_info.is_valid()) // it was a book move
      {
         if (count > 0) return; // not at learning threshold
         last_book_move = i;
         break;
      }
      else
      {
         if (entry.score() < min_score)
            min_score = entry.score();
         if (entry.score() > max_score)
            max_score = entry.score();
      }
      //if (count <0) return; // too far out of book
   }
   if (last_book_move < 0) return;
   int last_score = (*theLog)[theLog->current()-1].score();
   int last_depth = (*theLog)[theLog->current()-1].depth();
   if ((last_depth < LEARNING_MIN_DEPTH) ||
       (Util::Abs(last_score) < LEARNING_SCORE_THRESHOLD))
   {
      return; // we don't have an accurate score, or the
              // most recent score is close to even
   }
   theLog->write("learning"); theLog->write_eol();
   float learn_factor = (last_score*1.0F)/64.0F;
   BOOL first = TRUE;
   for (i = last_book_move; i >= 0; i-=2)
   {
      const Log_Entry &entry = (*theLog)[i];
      const Book_Info &book_info = entry.get_book_info();
      if (book_info.is_valid())
      {
         if (!first)
         {
            learn_factor /= 1.0F*book_info.get_total_moves();
         }
         update(book_info,learn_factor);
         first = FALSE;
      }
   }
}

void Book::update(const Book_Info &info, float learn_factor)
{
   Book_Location loc;
   info.get_location(loc);
   static char msg[100];

   sprintf(msg,"page = %d index = %d factor = %7.2f\n",loc.page,
           loc.index,learn_factor);
   theLog->write(msg);

   reader->update(loc,learn_factor);
}


