// Copyright 1993-1996 by Jon Dart.  All Rights Reserved.

#include "stdafx.h"
#include "bookread.h"
#include "constant.h"
#include "bhash.h"
#include "bookdefs.h"
#include "movegen.h"
#include "moveord.h"
#include "globals.h"
#include "util.h"
#include "config.h"
#include <string.h>
#include <iostream.h> // for debugging

static void show_error(const char *fileName)
{
   LPVOID errMsg;

   FormatMessage(
                FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                NULL,
                GetLastError(),
                MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                (LPTSTR)&errMsg,
                0,
                NULL 
                );
   static char *prefix = "Cannot open ";
   int len = strlen((LPCSTR)errMsg) + strlen(fileName) + strlen(prefix) + 2;

   char *msg = new char[len];
   strcpy(msg,prefix);
   strcat(msg,fileName);
   strcat(msg," ");
   strcat(msg,(LPCSTR)errMsg);
   MessageBox(NULL,msg,"Warning",
              MB_ICONEXCLAMATION | MB_OK);
   LocalFree(errMsg);
   delete [] msg;
}

Book_Reader::Book_Reader(LPCSTR fileName, LPCSTR mappingName)
: book_file(INVALID_HANDLE_VALUE),
pBook(NULL),
pPage(NULL),
hFileMapping(NULL),
open(FALSE),
current_page(-1)
{                 

   static char book_path[MAX_PATH];
   strcpy(book_path,programPath);
   {
      // find the last part of the program pathname
      char *p = strrchr(book_path,'\\');
      if (p)
      {
         // replace it with the book file name
         strcpy(p+1,fileName);
      }
      else
         strcpy(book_path,fileName);
      book_file = CreateFile(book_path,
#ifdef LEARNING
                             GENERIC_READ | GENERIC_WRITE,
                             FILE_SHARE_READ | FILE_SHARE_WRITE,
#else
                             GENERIC_READ,
                             FILE_SHARE_READ,
#endif                             
                             NULL /* security attributes */,
                             OPEN_EXISTING,
                             0,0);
      if (book_file == INVALID_HANDLE_VALUE)
      {
         show_error(fileName);
         return;
      }
   }
   if (book_file != INVALID_HANDLE_VALUE)
   {
      hFileMapping = CreateFileMapping(book_file,
                                       NULL,
#ifdef LEARNING
                                       PAGE_READWRITE,
#else
                                       PAGE_READONLY,
#endif
                                       0,
                                       0,
                                       NULL /*mappingName*/);
      if (!hFileMapping)
      {
         show_error(fileName);
         return;
      }
   }
   if (hFileMapping)
   {
      pBook = MapViewOfFile(hFileMapping,
#ifdef LEARNING
                            FILE_MAP_ALL_ACCESS,
#else
                            FILE_MAP_READ,
#endif
                            0,0,4096);
      if (!pBook)
      {
         show_error(fileName);
         return;
      }
   }

   open = pBook != NULL;
   if (!open)
      return;

   //memcpy(&hdr,(char*)pBook,Header_Size);
   char *p = (char*)pBook;
   hdr.version = (byte)(*p);
   hdr.num_pages = (byte)p[1];
   memcpy(&hdr.page_capacity,p+2,4);
   memcpy(&hdr.hash_table_size,p+6,4);

   if (hdr.version != (byte)Book_Version)
   {
      char msg[128];
      sprintf(msg,"Expected book version %d, got %d instead",
              (int)Book_Version,(int)hdr.version);
      MessageBox(NULL,msg,"Warning",MB_ICONEXCLAMATION);
      UnmapViewOfFile(pBook);
      CloseHandle(hFileMapping);
      CloseHandle(book_file);
      open = FALSE;
      return;
   }
   UnmapViewOfFile(pBook);
   pBook = NULL;
}

Book_Reader::~Book_Reader()
{
   if (pBook)
      UnmapViewOfFile(pBook);
   if (hFileMapping)
      CloseHandle(hFileMapping);
   if (book_file)
      CloseHandle(book_file);
}

const void Book_Reader::head( const Board & b, enum styles style,
                              Book_Location &loc)
{
   loc.page = (int)(b.HashCode() % hdr.num_pages);
   loc.index = INVALID;
   fetch_page(loc.page);
   unsigned probe = (unsigned)((b.HashCode()>>32) % hdr.hash_table_size);
   if (style != Balanced)
   {
      // Make the low-level bits of the bucket index match the
      // style
      probe &= ~(MAX_STYLES-1);
      probe |= (int)style;
   }

   // Copy the index from the book
   memcpy(&loc.index,pPage+Header_Size+sizeof(uint16)*probe,sizeof(uint16));
}

void Book_Reader::fetch( const Book_Location &loc, Book_Info &book_info )
{
   ASSERT(pPage);

   byte *entry = pPage + Header_Size + hdr.hash_table_size*2 + ((int)loc.index)*Entry_Size;
   hash_t hc;
   byte freq, rec, indx;
   // We assume here that the compiler doesn't reorder structures

   memcpy(&hc,entry,sizeof(hash_t));
   freq = entry[8];
   indx = entry[9];
   byte flags = entry[10];
   rec = flags & 0x7;
   book_info.init(hc,freq,flags,indx);
   book_info.winloss = entry[11];
   book_info.set_location(loc);
   memcpy(&(book_info.learn_score),entry+12,4);
}

void Book_Reader::update(const Book_Location &loc, float learn_factor)
{
#ifdef LEARNING
   fetch_page(loc.page);
   byte *entry = pPage + Header_Size + hdr.hash_table_size*2 + ((int)loc.index)*Entry_Size;
   // get the existing learn factor
   float old_factor;
   memcpy(&old_factor,entry+12,4);
   learn_factor += old_factor;
   // update the learn factor
   memcpy(entry+12,&learn_factor,4);
#endif
}

static int get_weight(const Book_Info &be,int total_freq,const ColorType side)
{
   int rec = be.get_recommend();
   if (rec != 50)
      return rec; // weight from the old book
   int w = be.get_winloss();
   int freqWeight = (int)((100L*be.get_frequency())/total_freq);
   // Slightly inflate the frequency of rare moves
   if (freqWeight < 10)
      freqWeight = freqWeight + (10-freqWeight)/4;
//     freqWeight = (freqWeight*(200-freqWeight))/100;
   TRACE("weight = %d\n",
         freqWeight);
   int winWeight = (w+100)/2;
   // Favor more frequent moves and moves that win
   int base = (freqWeight*winWeight)/200;
   if (base == 0) return 0;
   // Factor in learning
#ifdef LEARNING
   int score = (int)be.learn_score;
   if (score >= 0)
      return Util::Min(100,(base*(10+score))/10);
   else
      return Util::Max(0,(base*(10+score))/10);
#else
   return base;
#endif
}

Move Book_Reader::pick( const Board &b, const Book_Location &loc,
                        Book_Info &info )
{
   unsigned total_weight = 0;
   Book_Location tmp = loc;
   int total_freq = 0;
   int count = 0;
   // get the total move frequency
   while (tmp.index != INVALID)
   {
      Book_Info be;
      fetch(tmp, be);
      if (be.hash_code() == b.HashCode() && be.get_recommend() != 0)
      {
         total_freq += be.get_frequency();
         ++count;
      }
      if (be.is_last())
         break;
      else
         tmp.index++;
   }
   tmp = loc;
   // Determine the total weights of moves for this position.
   while (tmp.index != INVALID)
   {
      Book_Info be;
      fetch(tmp, be);
      if (b.HashCode() == be.hash_code())
      {
         int w  = get_weight(be,total_freq,b.Side());
         total_weight += w;
      }
      if (be.is_last())
         break;
      else
         tmp.index++;
   }
   
   // If total_weight is 0, no moves have non-zero weights.
   // Return without any move selected.
   if (total_weight == 0)
      return NullMove;
   unsigned weight = 0;
   tmp = loc;
   const unsigned nRand = (rand()*(unsigned long)total_weight)/RAND_MAX;
   // Randomly pick from the available moves.  Prefer moves
   // with high weights.
   while (tmp.index != INVALID)
   {
      Book_Info be;
      fetch(tmp,be);
      int w;
      if (b.HashCode() == be.hash_code() && (w = get_weight(be,total_freq,b.Side())) > 0)
      {
         weight += w;
         if (nRand < weight)
         {
            // We have selected a move. The book contains
            // only the move index, not the move itself.
            // We must call the move generator to obtain the
            // move.
            Move moves[Move_Generator::MaxMoves];
            Move_Generator mg( b, 0, NullMove);
            int n = mg.Generate_Moves(moves,TRUE /*repeatable*/); 
            ASSERT(be.move_index < n);
            be.set_move(moves[be.move_index]);
            info = be;
            info.set_total_moves(count);
            info.set_location(tmp);
            return be.get_move();
         }
      }
      if (be.is_last())
         break;
      else
         tmp.index++;
   }
   // should never get here
   ASSERT(FALSE);
   return NullMove;
}

int Book_Reader::book_moves(const Board &b, Move *moves, const unsigned limit)
{
   Book_Location tmp;
   head(b,Balanced,tmp);
   Book_Entry target(b.HashCode(),0,0,0);
   int num_moves = 0;
   int total_freq = 0;
   // get the total move frequency
   while (tmp.index != INVALID)
   {
      Book_Info be;
      fetch(tmp,be);
      if (target.is_equal(be) && be.get_recommend())
      {
         num_moves++;
         total_freq += be.get_frequency();
      }
      if (be.is_last())
         break;
      else
         tmp.index++;
   }
   if (num_moves == 0)
      return 0;
   Move *tmp_moves = new Move[Move_Generator::MaxMoves];
   Move_Generator mg( b, 0, NullMove);
   int n = mg.Generate_Moves(tmp_moves,TRUE /* repeatable */); 
   int * scores = new int[num_moves];
   Move * tmp_moves2 = new Move[num_moves];
   Book_Location loc;
   head(b,Balanced,loc);
   num_moves = 0;
   tmp = loc;
   while (tmp.index != INVALID)
   {
      Book_Info be;
      fetch(tmp,be);
      if (target.is_equal(be) && be.get_recommend())
      {
         ASSERT(be.move_index < n);
         tmp_moves2[num_moves] = tmp_moves[be.move_index];
         scores[num_moves] = get_weight(be,total_freq,b.Side());
         ++num_moves;
      }
      if (be.is_last())
         break;
      else
         tmp.index++;
   }
   Move_Ordering::sort_moves(tmp_moves2,scores,num_moves);
   unsigned ret_val = Util::Min(num_moves,limit);
   for (unsigned i = 0; i < ret_val; i++)
      moves[i] = tmp_moves2[i];
   delete [] scores;
   delete [] tmp_moves;
   delete [] tmp_moves2;
   return ret_val;
}

void Book_Reader::fetch_page(int page)
{
   if (current_page != page)
   {
      if (pBook)
      {
         UnmapViewOfFile(pBook);
      }
      pBook = MapViewOfFile(hFileMapping,
#ifdef LEARNING
                            FILE_MAP_ALL_ACCESS,
#else
                            FILE_MAP_READ,
#endif
                            0,page*hdr.page_capacity,hdr.page_capacity);
      if (!pBook)
      {
         MessageBox( NULL, "File mapping failed for opening book", "Warning", MB_ICONEXCLAMATION);
         pPage = NULL;
         return;
      }
      pPage = (byte*)pBook;
   }
   current_page = page;
}
