/* Copyright 1999 Jon Dart. All Rights Reserved. */
/* -*-mode: C++-*- */

#include "stdafx.h"
#include "hash.h"
#include "util.h"
#include "debug.h"
#include <malloc.h>
#include <memory.h>
#include <stddef.h>

static Position_Info *hashTable = NULL;
static Position_Info *rhashTable = NULL;
static int hashSize = 280000;
static int rehashSize;
static const int MaxRehash = 6;
static int refCount = 0;
#ifdef _WINDOWS
static CRITICAL_SECTION hashCritSection;
#endif

void initHash(int size)
{
#ifdef _WINDOWS
   InitializeCriticalSection(&hashCritSection);
   EnterCriticalSection(&hashCritSection);
#endif
   hashSize = size/2;
   rehashSize = hashSize/128;
   hashTable = (Position_Info*)malloc(sizeof(Position_Info)*hashSize);
   rhashTable = (Position_Info*)malloc(sizeof(Position_Info)*hashSize);
   memset(hashTable,'\0',sizeof(Position_Info)*hashSize);
   memset(rhashTable,'\0',sizeof(Position_Info)*hashSize);
   ++refCount;
#ifdef _WINDOWS
   LeaveCriticalSection(&hashCritSection);
#endif   
}

void freeHash()
{
   if (--refCount == 0)
   {
     free(hashTable);
     free(rhashTable);
#ifdef _WINDOWS
     DeleteCriticalSection(&hashCritSection);
#endif
   }
}

Position_Info *searchHash(const Board &currentBoard, int rep_count)
{
#ifdef _WINDOWS
   EnterCriticalSection(&hashCritSection);
#endif
   hash_t probe = quickmod(currentBoard.HashCode(),hashSize);
   Position_Info look(currentBoard,rep_count);
   Position_Info *p = (Position_Info*)&(rhashTable[probe]);
   if (p->is_equal(look)) 
   {
#ifdef _WINDOWS
      LeaveCriticalSection(&hashCritSection);
#endif
      return p;
   }
   int rehash = Bitmap(currentBoard.HashCode()).hivalue() % rehashSize;
   for (int i = 0; i < MaxRehash; i++)
   {
      p = (Position_Info*)&(hashTable[probe]);
      if (p->is_equal(look))
      {
#ifdef _WINDOWS
          LeaveCriticalSection(&hashCritSection);
#endif
     	  return p;
      }
      probe = (probe + rehash) % hashSize;
   }
#ifdef _WINDOWS
   LeaveCriticalSection(&hashCritSection);
#endif   
   return NULL;
}

int storeHash(const Board &currentBoard,
 int value, int alpha, int beta, Position_Info &newPI)
{
#ifdef _WINDOWS
   EnterCriticalSection(&hashCritSection);
#endif   
   int probe = quickmod(currentBoard.HashCode(),hashSize);
   rhashTable[probe] = newPI;
   int rehash = Bitmap(currentBoard.HashCode()).hivalue() % rehashSize;
   int best = -1;
   int minDepth = 10000;
   int minProbe = -1;
   int retVal = 0;
   for (int i = 0; i < MaxRehash; i++)
   {
      Position_Info *p = (Position_Info*)&hashTable[probe];
      
      if (p->hash_code() == (hash_t)0)
      {
         best = probe;
         break;
      }
      if (p->age() != newPI.age())
      {
         best = probe;
         break;
      }
      else if (p->depth() < minDepth)
      {
         minProbe = probe; minDepth = p->depth();
      }
      probe = (probe + rehash) % hashSize;
   }
   if (best == -1 && newPI.depth() > minDepth)
   {
      // replacing
      best = minProbe;
      retVal = 1;
   }
   if (best != -1)
   {
      UNCOND_ASSERT(best >= 0 && best < hashSize);
      hashTable[best] = newPI;
   }
   else
   {
#ifdef _WINDOWS
      LeaveCriticalSection(&hashCritSection);
#endif   
      return 2; /* fail*/
   }
#ifdef _WINDOWS
   LeaveCriticalSection(&hashCritSection);
#endif   
   return retVal;
      
}

Hash::Hash(const unsigned size, 
           const unsigned long max_entries)
{
   my_size = size;
   my_max_entries = max_entries;
   hash_entries = 0L;
   ht = new HashList * [my_size];
   ASSERT(ht);
   for (unsigned i=0;i<my_size;i++)
      ht[i] = NULL;
}


Hash::~Hash()
{
   clear();
   delete [] ht;
}

Hashable * Hash::search( const Hashable &to_match ) const
{
   unsigned probe = (unsigned)(quickmod(to_match.hash_code(),my_size));
   UNCOND_ASSERT(probe < my_size);
   HashList *entry = ht[probe];
   if (!entry)
      return NULL;
   POSITION pos = entry->GetHeadPosition();
   while (pos)
   {
      Hashable *c = (Hashable*)entry->GetNext(pos);
      if (c->is_equal(to_match))
         return c;
   }
   return NULL;
}

BOOL Hash::insert(Hashable *p, BOOL checkPresent)
{
   if (hash_entries >= my_max_entries)
   {
      return FALSE;
   }
   hash_t h = p->hash_code();
   unsigned probe = (unsigned)(quickmod(h,my_size));
   UNCOND_ASSERT(probe < my_size);
   HashList *entry = ht[probe];
   if (!entry)
   {
      ht[probe] = new HashList;
      ASSERT(ht[probe]);
      ht[probe]->AddTail(p);
   }
   else
   {
      if (checkPresent)
      {
         POSITION pos = entry->GetHeadPosition();
         while (pos)
         {
            Hashable *c = (Hashable*)entry->GetNext(pos);
            if (c->is_equal(*p))
               return FALSE; // already in there
         }
      }
      entry->AddHead(p);
   }
   hash_entries++;
   return TRUE;
}

void Hash::remove(Hashable *p)
{
   hash_t h = p->hash_code();
   unsigned probe = (unsigned)(quickmod(h,my_size));
   UNCOND_ASSERT(probe < my_size);
   HashList *entry = ht[probe];
   if (entry)
   {
      POSITION pos = entry->GetHeadPosition();
      while (pos)
      {
         POSITION prev = pos;
         Hashable *c = (Hashable*)entry->GetNext(pos);
         if (c->is_equal(*p))
         {
            entry->RemoveAt(prev);
            delete c;
            hash_entries--;
            return;
         }
      }
   }
}

void Hash::clear(BOOL final)
{
   for (unsigned i = 0; i < my_size; i++)
   {
      if (ht[i])
      {
         HashList *list = ht[i];
         delete list;
         ht[i] = NULL;
      }
   }
   hash_entries = 0;
}

unsigned long Hash::num_entries() const
{
   return hash_entries;
}





